// 3DS Test program
// Portions copyright IBM Corporation
// Portions copyright Robert Basler, 1998

//#define  INCL_PM                    /* required to use PM APIs.            */
#define  INCL_OS2MM                 /* required for MCI and MMIO headers   */
#define  INCL_DOS

#include <os2.h>
extern "C" {
#include <os2me.h>
};
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <process.h>
#include <conio.h>
#include "sems.h"
#include "3dsound.h"
#include "3dstestr.h"
#include "resource.h"

// Length of MCI error message result string
#define  STRING_LENGTH  128
// Number of buffers for playing WAV file.
#define  MAX_BUFFERS    32


USHORT               usDeviceID;                /* Amp Mixer device id     */
ULONG                ulBufferHead;              /* Current file buffer     */
ULONG                ulBufferTail;              /* Current file buffer tail */
ULONG                ulNumBuffers;              /* Number of file buffers  */

MCI_MIX_BUFFER       MixBuffers[MAX_BUFFERS];   /* Device buffers          */
MCI_MIXSETUP_PARMS   MixSetupParms;             /* Mixer parameters        */
MCI_BUFFER_PARMS     BufferParms;               /* Device buffer parms     */

TEvSemHandle hMixSem;
TMxSemHandle hQSem;
HQUEUE QHandle;

// Parameters used to set up the mixer
ULONG uBitsPerSample = 16, uSamplesPerSec = 11025, uChannels = 2,
      uMixBufferSize = 0, uSurroundSound = 0, uSwapLeftRight = 0,
      uMaxTracks = MAXTRACKS;

uchar AudioResourceName[256] = "3dstest.res";



/***************************************************************************
 * Name        : MciError
 *
 * Description : Display a message box with the string corresponding to
 *               the passed mci error code.
 *
 * Parameters  : ulError = mci error code
 *
 * Return      : None
 *
 ***************************************************************************/
VOID MciError( ULONG ulError )
{
   SZ       szBuffer[ STRING_LENGTH ];
   ULONG    rc;
   rc = mciGetErrorString( ulError, szBuffer, STRING_LENGTH );
   printf("S3D Sound MCI Error: %s\n",szBuffer);
}

/***************************************************************************
 * Name        : APIError
 *
 * Description : Display a message box with the string provided
 *
 * Parameters  : ulError = mci error code
 *
 * Return      : None
 *
 ***************************************************************************/
VOID APIError(APIRET Error, uchar *msg )
{
   printf("S3D Sound Error: %s\nRetCode: %i\n",msg,Error);
}


/***************************************************************************
 * Name        : MyEvent
 *
 * Description : The address to this procedure is passed to the mixer
 *               device in the MIX_SETUP_PARMS structure. The mixer
 *               device then calls this procedure when it has expended
 *               a buffer.
 *
 *               NOTE: This is a high priority thread. Too much code here
 *                     will bog the system down.
 *
 * Parameters  : ulStatus - Detailed error message
 *               pBuffer  - Pointer to expended buffer
 *               ulFlags  - Indicates the type of event
 *
 *
 * Return      : BOOL   - TRUE  = failure
 *                      - FALSE = success
 *
 ***************************************************************************/

LONG APIENTRY MyEvent ( ULONG ulStatus,
                        PMCI_MIX_BUFFER  pBuffer,
                        ULONG  ulFlags  )

{

   switch( ulFlags )
   {
   case MIX_STREAM_ERROR | MIX_READ_COMPLETE :  /* error occur in device */
   case MIX_STREAM_ERROR | MIX_WRITE_COMPLETE:  /* error occur in device */

      if ( ulStatus == ERROR_DEVICE_UNDERRUN)
      {
         /* handle ERROR_DEVICE_UNDERRUN here
          */
//         WinPostMsg( hwndFrame, WM_COMMAND, ( PVOID )ID_STOP, 0L );
      }
      if ( ulStatus == ERROR_DEVICE_OVERRUN)
      {
         /* handle ERROR_DEVICE_OVERRUN here
          */
//         WinPostMsg( hwndFrame, WM_COMMAND, ( PVOID )ID_STOP, 0L );
      }

      break;

   case MIX_WRITE_COMPLETE:           /* for playback  */
        // And next buffer to the sound device (mixer)
        MxSemRequest(hQSem,0);
        if (ulBufferTail!=ulBufferHead) {
             ulBufferTail=(ulBufferTail+1)%ulNumBuffers;
             EvSemPost(hMixSem);
        } 
        else {
             ulBufferTail=(ulBufferTail+1)%ulNumBuffers;
             memset(MixBuffers[ulBufferTail].pBuffer,0,uMixBufferSize);
        };
        MxSemRelease(hQSem);
//printf("%i Played.\n",ulBufferTail);
        MixSetupParms.pmixWrite( MixSetupParms.ulMixHandle,
                                 &MixBuffers[ulBufferTail],
                                 1 );
        break;

   } /* end switch */
   return( TRUE );
} /* end MyEvent */


void MixingThread(void*) {
int count;
//printf("Mixing thread started.\n");
//DosSetPriority(PRTYS_THREAD,PRTYC_REGULAR,28,0);
do {
   count=EvSemWait(hMixSem,1000);
   while (count>0) {
        // Get buffer from mixing library
        MxSemRequest(hQSem,0);
        ulBufferHead=(ulBufferHead+1)%ulNumBuffers;
        MxSemRelease(hQSem);
        S3DI_NextBuffer(MixBuffers[ulBufferHead].pBuffer);
//printf("%i Mixed.\n",ulBufferHead);
        count--;
   } /* endwhile */
} while (1); /* enddo */
_endthread();
};


VOID _Optlink QMonitor( PVOID qh)
 {
  APIRET      rc=0;
  BYTE        Priority;
  REQUESTDATA RD;
  ULONG       mp1, mp2;
 
//  printf("QMonitor Started!\n");
  while (1)
   {
    DosReadQueue((HQUEUE)qh, &RD, (PULONG)&mp1, (PPVOID)&mp2, 0L, (BOOL32)0,
                 &Priority, (HEV)0);
    switch(RD.ulData)
      {
        case MM_MCINOTIFY:
           printf("   msg=MM_MCINOTIFY:\n");
           printf("   mp1=%d\n",mp1);
           printf("   mp2=%d\n",mp2);
           break;
        case MM_MCIPASSDEVICE:
           printf("   msg=MM_MCIPASSDEVICE: Device id=%d ",mp1);
           switch (mp2)
             {
               case MCI_LOSING_USE:
                  printf("MCI_LOSING_USE\n");
                  break;
               case MCI_GAINING_USE:
                  printf("MCI_GAINING_USE\n");
                  break;
               default:
                  printf("mp2=%d\n",mp2);
                  break;
             }
           break;
        case MM_MCIPOSITIONCHANGE:
           printf("   msg=MM_MCIPOSITIONCHANGE:\n");
           printf("   mp1=%d\n",mp1);
           printf("   mp2=%d\n",mp2);
           break;
        case MM_MCICUEPOINT:
           printf("   msg=MM_MCICUEPOINT:\n");
           printf("   mp1=%d\n",mp1);
           printf("   mp2=%d\n",mp2);
           break;
        case MM_MCIPLAYLISTMESSAGE:
           printf("   msg=MM_MCIPLAYLISTMESSAGE:\n");
           printf("   mp1=%d\n",mp1);
           printf("   mp2=%d\n",mp2);
           break;
        case MM_MCIEVENT:
           printf("   msg=MM_MCIEVENT:\n");
           printf("   mp1=%d\n",mp1);
           printf("   mp2=%d\n",mp2);
           break;
        default:
           printf("   msg=%d\n",RD.ulData);
           printf("   mp1=%d\n",mp1);
           printf("   mp2=%d\n",mp2);
           break;
      }
   }   /* endwhile */
  return;
 }



/***************************************************************************
 * Name        : Close
 *
 * Description : Deallocate memory and close the Amp-Mixer Device.
 *
 *
 * Parameters  : None
 *
 * Return      : BOOL   - TRUE  = failure
 *                      - FALSE = success
 *
 ***************************************************************************/
VOID Close( VOID )
{
   MCI_GENERIC_PARMS    GenericParms;
   ULONG                rc;

   rc = mciSendCommand( usDeviceID,
                        MCI_BUFFER,
                        MCI_WAIT | MCI_DEALLOCATE_MEMORY,
                        ( PVOID )&BufferParms,
                        0 );

   if ( rc != MCIERR_SUCCESS )
   {
      MciError( rc );
      return;
   }

   rc = mciSendCommand( usDeviceID,
                        MCI_MIXSETUP,
                        MCI_WAIT | MCI_MIXSETUP_DEINIT,
                        ( PVOID ) &MixSetupParms,
                        0 );

   if ( rc != MCIERR_SUCCESS )
   {
      MciError( rc );
      return;
   }

   rc = mciSendCommand( usDeviceID,
                        MCI_CLOSE,
                        MCI_WAIT ,
                        ( PVOID )&GenericParms,
                        0 );

   if ( rc != MCIERR_SUCCESS )
   {
      MciError( rc );
      return;
   }

  DosCloseQueue(QHandle);
}



/***************************************************************************
 * Name        : SetupAmpmix
 *
 * Description : Allocate application buffers for audio file and load
 *               audio file. These buffers will be copied to device
 *               buffers by another thread.
 *
 *               Before this procedure loads the audio file, the global
 *               MixSetupParms data structure is loaded with information
 *               from the audio files header.
 *
 *
 * Return      : BOOL   - TRUE  = failure
 *                      - FALSE = success
 *
 ***************************************************************************/
BOOL SetupAmpmix(void)
{
   MCI_AMP_OPEN_PARMS   AmpOpenParms;
   ULONG                rc, ulIndex;

  PID      OwnerPID=0;
  HQUEUE   QHandle =0;
  char     QueueName[100];
 
  strcpy(QueueName, "\\QUEUES\\S3DQUEUE");
  if (DosCreateQueue(&QHandle, QUE_FIFO, QueueName)) return(1L);
  if (DosOpenQueue(&OwnerPID, &QHandle, QueueName))  return(1L);
  _beginthread((void(*)(void*))QMonitor, (PVOID)NULL, 65536L, (PVOID)QHandle);

   /* open the mixer device
    */
   memset ( &AmpOpenParms, '\0', sizeof ( MCI_AMP_OPEN_PARMS ) );
   AmpOpenParms.usDeviceID = ( USHORT ) 0;
   AmpOpenParms.pszDeviceType = ( PSZ ) MCI_DEVTYPE_AUDIO_AMPMIX;
   AmpOpenParms.hwndCallback = QHandle;

   rc = mciSendCommand( 0,
                       MCI_OPEN,
                       MCI_DOS_QUEUE | MCI_WAIT | MCI_OPEN_TYPE_ID | MCI_OPEN_SHAREABLE,
                       ( PVOID ) &AmpOpenParms,
                       0 );

   if ( rc != MCIERR_SUCCESS )
   {
      MciError( rc );
      return( TRUE );
   }

   usDeviceID = AmpOpenParms.usDeviceID;

   /* Set the MixSetupParms data structure to match the loaded file.
    * This is a global that is used to setup the mixer.
    */
   memset( &MixSetupParms, '\0', sizeof( MCI_MIXSETUP_PARMS ) );

   MixSetupParms.ulBitsPerSample = uBitsPerSample;
   MixSetupParms.ulFormatTag = MCI_WAVE_FORMAT_PCM;
   MixSetupParms.ulSamplesPerSec = uSamplesPerSec;
   MixSetupParms.ulChannels = uChannels;

   /* Setup the mixer for playback of wave data
    */
   MixSetupParms.ulFormatMode = MCI_PLAY;
   MixSetupParms.ulDeviceType = MCI_DEVTYPE_WAVEFORM_AUDIO;
   MixSetupParms.pmixEvent    = MyEvent;

   rc = mciSendCommand( usDeviceID,
                        MCI_MIXSETUP,
                        MCI_WAIT | MCI_MIXSETUP_INIT,
                        ( PVOID ) &MixSetupParms,
                        0 );

   if ( rc != MCIERR_SUCCESS )
   {
      MciError( rc );
      return( TRUE );
   }

   /* Use the suggested buffer size provide by the mixer device
    * and 5 seconds of audio to calculate the required
    * number of Amp-Mixer buffers.
    */
   uMixBufferSize=MixSetupParms.ulBufferSize;
   ulNumBuffers = (5 * uSamplesPerSec * (uBitsPerSample/8) *uChannels) / uMixBufferSize +1;
   if (ulNumBuffers>MAX_BUFFERS) ulNumBuffers=MAX_BUFFERS;


   /* Set up the BufferParms data structure and allocate
    * device buffers from the Amp-Mixer
    */
   BufferParms.ulStructLength = sizeof(MCI_BUFFER_PARMS);
   BufferParms.ulNumBuffers = ulNumBuffers;
   BufferParms.ulBufferSize = MixSetupParms.ulBufferSize;
   BufferParms.pBufList = MixBuffers;

   rc = mciSendCommand( usDeviceID,
                        MCI_BUFFER,
                        MCI_WAIT | MCI_ALLOCATE_MEMORY,
                        ( PVOID ) &BufferParms,
                        0 );

    if ( ULONG_LOWD( rc) != MCIERR_SUCCESS )
    {
      MciError( rc );
      return( TRUE );
    }

   /* Set up all device buffers
    */
   for( ulIndex = 0; ulIndex < ulNumBuffers; ulIndex++)
   {
      memset( MixBuffers[ ulIndex ].pBuffer, '\0', BufferParms.ulBufferSize );
      MixBuffers[ ulIndex ].ulBufferLength = BufferParms.ulBufferSize;
   }

   /* Set the "end-of-stream" flag
    */
//   MixBuffers[ulNumBuffers - 1].ulFlags = MIX_BUFFER_EOS;

   return( FALSE );

}

void main(int argc, char *argv[]) {
int16 rc, ch;
uchar buf[100];
if (argc>1) {
     Resource r("3dstest.res");
     r.Build("3dstestr.h");
     return;
} /* endif */
printf("Stereo or Mono (S/m)\n");
ch=getch();
if ((ch==0x0d)||(ch=='S')||(ch=='s')) uChannels=2;
else uChannels=1;
printf("Normal or Surround (N/s)\n");
ch=getch();
if ((ch==0x0d)||(ch=='N')||(ch=='n')) uSurroundSound=0;
else uSurroundSound=1;
printf("16 or 8-bit output (1/8)\n");
ch=getch();
if ((ch==0x0d)||(ch=='1')) uBitsPerSample=16;
else uBitsPerSample=8;
printf("Swap channels (N/y)\n");
ch=getch();
if ((ch==0x0d)||(ch=='N')) uSwapLeftRight=0;
else uSwapLeftRight=1;
/* Load the audio file and setup for playback. */
rc = SetupAmpmix();
if (rc!=0) APIError(0,(uchar*)"SetupAmpmix");
printf("Setting mixer for\n %i bits per sample\n %i Hz\n %i Channels\n %i byte buffer\n Surround is %i\n Swap Channels is %i\n %i tracks max.\n",
     uBitsPerSample,uSamplesPerSec,uChannels,uMixBufferSize,uSurroundSound,uSwapLeftRight,
      uMaxTracks);

rc = S3D_Init(AudioResourceName,0.0,uMixBufferSize,uBitsPerSample,uSamplesPerSec,uChannels,uSurroundSound,uSwapLeftRight);
if (rc!=S3DE_NONE) APIError(rc,S3D_ErrorMessage(rc,buf,100));
S3D_SetMaxTracks(uMaxTracks);
EvSemCreate(&hMixSem,0);
MxSemCreate(&hQSem,0);
ulBufferHead=ulBufferTail=0;
// Get three buffers initially.
ulBufferHead=(ulBufferHead+1)%ulNumBuffers;
S3DI_NextBuffer(MixBuffers[ulBufferHead].pBuffer);
ulBufferHead=(ulBufferHead+1)%ulNumBuffers;
S3DI_NextBuffer(MixBuffers[ulBufferHead].pBuffer);
ulBufferHead=(ulBufferHead+1)%ulNumBuffers;
S3DI_NextBuffer(MixBuffers[ulBufferHead].pBuffer);

/* Setup mixing thread */
_beginthread((void(*)(void*))MixingThread,NULL,32768,NULL);

// Start the amp-mix device playing
MxSemRequest(hQSem,0);
ulBufferTail=(ulBufferTail+1)%ulNumBuffers;
MxSemRelease(hQSem);
EvSemPost(hMixSem);

// Must write at least two buffers to start mixer
MixSetupParms.pmixWrite( MixSetupParms.ulMixHandle,
                         &MixBuffers[ulBufferTail],
                         2 );
rc=S3D_Volume(MAXVOLUME/2);
printf("S3D_Volume %s\n",S3D_ErrorMessage(rc,buf,100));

int volume=MAXVOLUME/2, v0=MAXVOLUME/2, v1=MAXVOLUME/2, v2=MAXVOLUME/2;
dword f;
if (rc==S3DE_NONE) {
     ch=' ';
     rc=1;
     f=0;
     while ((rc>=0) && (ch!=27)) {
          printf("Prefixes:\n");
          printf("L - play in left channel only.     ");
          printf("C - play in centre channel only.\n");
          printf("R - play in right channel only.    ");
          printf("S - play in surround channel only.\n");
          printf("Commands:\n");
          printf("1 - play 8-bit mono track.         \n");
          printf("2 - play 8-bit stereo track.       \n");
          printf("3 - play 16-bit mono track.\n");
          printf("4 - play 16-bit stereo track.\n");
          printf("5 - play 8-bit stereo track as stereo.\n");
          printf("6 - play 16-bit stereo track as stereo.\n");
          printf("Esc - Exit\n");
          ch=getch();
          switch(ch) {
               case 0x27 : break;
               case 'l' : 
               case 'L' : f=S3DTF_LEFT; break;
               case 'c' : 
               case 'C' : f=S3DTF_CENTRE; break;
               case 'r' : 
               case 'R' : f=S3DTF_RIGHT; break;
               case 's' : 
               case 'S' : f=S3DTF_SURROUND; break;
               case '1' : {
                    rc=S3D_TrackOpen(TESTWAVE8M,f+S3DTF_PLAYING+S3DTF_CLOSEAFTERPLAY,0,0,0,1,MAXVOLUME,4096);
                    if (rc>=0) printf("S3D_TrackOpen returned handle %i.\n",rc);
                    else printf("S3D_TrackOpen %s\n",S3D_ErrorMessage(rc,buf,100));
                    f=0;
                    break;
               };
               case '2' : {
                    rc=S3D_TrackOpen(TESTWAVE8S,f+S3DTF_PLAYING+S3DTF_CLOSEAFTERPLAY,0,0,0,1,MAXVOLUME,4096);
                    if (rc>=0) printf("S3D_TrackOpen returned handle %i.\n",rc);
                    else printf("S3D_TrackOpen %s\n",S3D_ErrorMessage(rc,buf,100));
                    f=0;
                    break;
               };
               case '3' : {
                    rc=S3D_TrackOpen(TESTWAVE16M,f+S3DTF_PLAYING+S3DTF_CLOSEAFTERPLAY,0,0,0,1,MAXVOLUME,4096);
                    if (rc>=0) printf("S3D_TrackOpen returned handle %i.\n",rc);
                    else printf("S3D_TrackOpen %s\n",S3D_ErrorMessage(rc,buf,100));
                    f=0;
                    break;
               };
               case '4' : {
                    rc=S3D_TrackOpen(TESTWAVE16S,f+S3DTF_PLAYING+S3DTF_CLOSEAFTERPLAY,0,0,0,1,MAXVOLUME,4096);
                    if (rc>=0) printf("S3D_TrackOpen returned handle %i.\n",rc);
                    else printf("S3D_TrackOpen %s\n",S3D_ErrorMessage(rc,buf,100));
                    f=0;
                    break;
               };
               case '5' : {
                    rc=S3D_TrackOpen(TESTWAVE8S,S3DTF_PLAYING+S3DTF_CLOSEAFTERPLAY+S3DTF_STEREOMIXING,0,0,0,1,MAXVOLUME,4096);
                    if (rc>=0) printf("S3D_TrackOpen returned handle %i.\n",rc);
                    else printf("S3D_TrackOpen %s\n",S3D_ErrorMessage(rc,buf,100));
                    break;
               };
               case '6' : {
                    rc=S3D_TrackOpen(TESTWAVE16S,S3DTF_PLAYING+S3DTF_CLOSEAFTERPLAY+S3DTF_STEREOMIXING,0,0,0,1,MAXVOLUME,4096);
                    if (rc>=0) printf("S3D_TrackOpen returned handle %i.\n",rc);
                    else printf("S3D_TrackOpen %s\n",S3D_ErrorMessage(rc,buf,100));
                    break;
               };
               default: {
                  switch(ch) {
                    case '.' : volume+=2; if (volume>MAXVOLUME) volume=MAXVOLUME; v0=v1=v2=volume; break;
                    case ',' : volume-=2; if (volume<0) volume=0; v0=v1=v2=volume; break;
                    case 'g' : v0-=2; if (v0<0) v0=0; break;
                    case 'j' : v1-=2; if (v1<0) v1=0; break;
                    case 'n' : v2-=2; if (v2<0) v2=0; break;
                    case 'h' : v0+=2; if (v0>255) v0=255; break;
                    case 'k' : v1+=2; if (v1>255) v1=255; break;
                    case 'm' : v2+=2; if (v2>255) v2=255; break;
                  };
                  if (volume>MAXVOLUME) volume=MAXVOLUME;
                  if (volume<0) volume=0;
                  printf("Vol %i Left %i Right %i Surround %i\n",volume,v0,v1,v2); 
                  S3D_Volume(volume); 
                  S3D_TrackVolumeAll(rc,v0,v1,v2);
               };
          };
     } /* endwhile */
};

printf("Shutting down...\n");
// Shut down the mixing library
if ((rc==S3DE_NONE) || (rc>=0)) {
     printf("Waiting for tracks to end.\n");
     S3D_TrackWait(S3DT_WAITALL);
     printf("Waiting for buffers to flush.\n");
     // Give buffers 5 seconds to flush
     DosSleep(5000);

     // Shut down the ampmix device
     Close();

     rc=S3D_Done();
     printf("S3D_Done %s\n",S3D_ErrorMessage(rc,buf,100));
};

EvSemClose(hMixSem);
MxSemClose(hQSem);

};

